@@ -27,6 +27,8 @@ from imdb import IMDb
 
 from functools import partial
 
+from unidecode import unidecode
+
 setproctitle.setproctitle("hypnotix")
 
 # i18n
@@ -96,6 +98,7 @@ class MainWindow():
         self.back_page = None # page to go back to if the back button is pressed
         self.active_channel = None
         self.fullscreen = False
+        self.latest_search_bar_text = None
         self.mpv = None
         self.ia = IMDb()
 
@@ -105,7 +108,9 @@ class MainWindow():
         self.x_pos = 0
         self.y_pos = 0
 
-        # Set the Glade file
+        # Used for redownloading timer
+        self.reload_timeout_sec = 60*5
+        self._timerid = -1
         gladefile = "/usr/share/hypnotix/hypnotix.ui"
         self.builder = Gtk.Builder()
         self.builder.set_translation_domain(APP)
@@ -131,9 +136,8 @@ class MainWindow():
         self.edit_mode = False
 
         # Create variables to quickly access dynamic widgets
-        self.generic_channel_pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size("/usr/share/hypnotix/generic_tv_logo.png", 22, 22)
-        widget_names = ["headerbar", "status_label", "status_bar", "sidebar", "go_back_button", "channels_box", \
-            "provider_button", "preferences_button", \
+        widget_names = ["headerbar", "status_label", "status_bar", "sidebar", "go_back_button", "search_button", "search_bar", \
+            "channels_box", "provider_button", "preferences_button", \
             "mpv_drawing_area", "stack", "fullscreen_button", \
             "provider_ok_button", "provider_cancel_button", \
             "name_entry", "path_label", "path_entry", "browse_button", "url_label", "url_entry", \
@@ -151,6 +155,7 @@ class MainWindow():
             "info_section", "info_revealer", "info_name_label", "info_plot_label", "info_rating_label", "info_year_label", "close_info_button", \
             "info_genre_label", "info_duration_label", "info_votes_label", "info_pg_label", "divider_label", \
             "useragent_entry", "referer_entry", "mpv_entry", "mpv_link", \
+            "darkmode_switch",
             "mpv_stack", "spinner", "info_window_close_button", \
             "video_properties_box", "video_properties_label", \
             "colour_properties_box", "colour_properties_label", \
@@ -188,6 +193,9 @@ class MainWindow():
         self.series_button.connect("clicked", self.show_groups, SERIES_GROUP)
         self.go_back_button.connect("clicked", self.on_go_back_button)
 
+        self.search_button.connect("toggled", self.on_search_button_toggled)
+        self.search_bar.connect("activate", self.on_search_bar)
+
         self.stop_button.connect("clicked", self.on_stop_button)
         self.pause_button.connect("clicked", self.on_pause_button)
         self.show_button.connect("clicked", self.on_show_button)
@@ -214,6 +222,12 @@ class MainWindow():
         self.bind_setting_widget("http-referer", self.referer_entry)
         self.bind_setting_widget("mpv-options", self.mpv_entry)
 
+        #dark mode
+        prefer_dark_mode = self.settings.get_boolean("prefer-dark-mode")
+        Gtk.Settings.get_default().set_property("gtk-application-prefer-dark-theme", prefer_dark_mode)
+        self.darkmode_switch.set_active(prefer_dark_mode)
+        self.darkmode_switch.connect("notify::active", self.on_darkmode_switch_toggled)
+
         # Menubar
         accel_group = Gtk.AccelGroup()
         self.window.add_accel_group(accel_group)
@@ -264,31 +278,47 @@ class MainWindow():
         self.provider_type_combo.set_active(0) # Select 1st type
         self.provider_type_combo.connect("changed", self.on_provider_type_combo_changed)
 
-        self.tv_logo.set_from_pixbuf(GdkPixbuf.Pixbuf.new_from_file_at_size("/usr/share/hypnotix/pictures/tv.svg", 258, 258))
-        self.movies_logo.set_from_pixbuf(GdkPixbuf.Pixbuf.new_from_file_at_size("/usr/share/hypnotix/pictures/movies.svg", 258, 258))
-        self.series_logo.set_from_pixbuf(GdkPixbuf.Pixbuf.new_from_file_at_size("/usr/share/hypnotix/pictures/series.svg", 258, 258))
+        self.tv_logo.set_from_surface(self.get_surface_for_file("/usr/share/hypnotix/pictures/tv.svg", 258, 258))
+        self.movies_logo.set_from_surface(self.get_surface_for_file("/usr/share/hypnotix/pictures/movies.svg", 258, 258))
+        self.series_logo.set_from_surface(self.get_surface_for_file("/usr/share/hypnotix/pictures/series.svg", 258, 258))
 
         self.reload(page="landing_page")
 
-        # Redownload playlists after a little while...
-        GLib.timeout_add_seconds(60 * 5, self.force_reload)
+        # Redownload playlists by default
+        # This is going to get readjusted
+        self._timerid = GLib.timeout_add_seconds(self.reload_timeout_sec, self.force_reload)
 
         self.window.show()
         self.playback_bar.hide()
+        self.search_bar.hide()
 
         # Historic bitrates of the currently playing media
         self.video_bitrates = []
         self.audio_bitrates = []
 
+
+    def get_surface_for_file(self, filename, width, height):
+        scale = self.window.get_scale_factor()
+        if width != -1:
+            width = width * scale
+        if height != -1:
+            height = height * scale
+
+        pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(filename, width, height)
+        surface = Gdk.cairo_surface_create_from_pixbuf(pixbuf, scale)
+        return surface
+
+    def get_surf_based_image(self, filename, width, height):
+        surf = self.get_surface_for_file(filename, width, height)
+        return Gtk.Image.new_from_surface(surf)
+
     def add_badge(self, word, box, added_words):
         if word not in added_words:
             for extension in ["svg", "png"]:
                 badge = "/usr/share/hypnotix/pictures/badges/%s.%s" % (word, extension)
                 if os.path.exists(badge):
                     try:
-                        pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(badge, -1, 16)
-                        surface = Gdk.cairo_surface_create_from_pixbuf(pixbuf, self.window.get_scale_factor())
-                        image = Gtk.Image().new_from_surface(surface)
+                        image = self.get_surf_based_image(badge, -1, 32)
                         box.pack_start(image, False, False, 0)
                         added_words.append(word)
                         break
@@ -365,9 +395,7 @@ class MainWindow():
                 label.set_max_width_chars(30)
                 label.set_ellipsize(Pango.EllipsizeMode.END)
                 box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
-                pixbuf = self.get_pixbuf(channel.logo_path)
-                surface = Gdk.cairo_surface_create_from_pixbuf(pixbuf, self.window.get_scale_factor())
-                image = Gtk.Image().new_from_surface(surface)
+                image = Gtk.Image().new_from_surface(self.get_channel_surface(channel.logo_path))
                 logos_to_refresh.append((channel, image))
                 box.pack_start(image, False, False, 0)
                 box.pack_start(label, False, False, 0)
@@ -396,9 +424,7 @@ class MainWindow():
             label.set_max_width_chars(30)
             label.set_ellipsize(Pango.EllipsizeMode.END)
             box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
-            pixbuf = self.get_pixbuf(item.logo_path)
-            surface = Gdk.cairo_surface_create_from_pixbuf(pixbuf, self.window.get_scale_factor())
-            image = Gtk.Image().new_from_surface(surface)
+            image = Gtk.Image().new_from_surface(self.get_channel_surface(item.logo_path))
             logos_to_refresh.append((item, image))
             box.pack_start(image, False, False, 0)
             box.pack_start(label, False, False, 0)
@@ -420,6 +446,11 @@ class MainWindow():
     def show_episodes(self, serie):
         logos_to_refresh = []
         self.active_serie = serie
+        # If we are using xtream provider
+        # Load every Episodes of every Season for this Series
+        if self.active_provider.type_id == "xtream":
+            self.x.get_series_info_by_id(self.active_serie)
+
         self.navigate_to("episodes_page")
         for child in self.episodes_box.get_children():
             self.episodes_box.remove(child)
@@ -440,9 +471,7 @@ class MainWindow():
                 label.set_max_width_chars(30)
                 label.set_ellipsize(Pango.EllipsizeMode.END)
                 box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
-                pixbuf = self.get_pixbuf(episode.logo_path)
-                surface = Gdk.cairo_surface_create_from_pixbuf(pixbuf, self.window.get_scale_factor())
-                image = Gtk.Image().new_from_surface(surface)
+                image = Gtk.Image().new_from_surface(self.get_channel_surface(episode.logo_path))
                 logos_to_refresh.append((episode, image))
                 box.pack_start(image, False, False, 0)
                 box.pack_start(label, False, False, 0)
@@ -474,6 +503,11 @@ class MainWindow():
     def on_entry_changed(self, widget, key):
         self.settings.set_string(key, widget.get_text())
 
+    def on_darkmode_switch_toggled(self, widget, key):
+        prefer_dark_mode = widget.get_active()
+        self.settings.set_boolean("prefer-dark-mode", prefer_dark_mode)
+        Gtk.Settings.get_default().set_property("gtk-application-prefer-dark-theme", prefer_dark_mode)
+
     @async_function
     def download_channel_logos(self, logos_to_refresh):
         headers = {
@@ -483,6 +517,8 @@ class MainWindow():
         for channel, image in logos_to_refresh:
             if channel.logo_path == None:
                 continue
+            if os.path.isfile(channel.logo_path):
+                continue
             try:
                 response = requests.get(channel.logo, headers=headers, timeout=10, stream=True)
                 if response.status_code == 200:
@@ -495,30 +531,71 @@ class MainWindow():
 
     @idle_function
     def refresh_channel_logo(self, channel, image):
-        pixbuf = self.get_pixbuf(channel.logo_path)
-        surface = Gdk.cairo_surface_create_from_pixbuf(pixbuf, self.window.get_scale_factor())
-        image.set_from_surface(surface)
+        image.set_from_surface(self.get_channel_surface(channel.logo_path))
 
-    def get_pixbuf(self, path):
+    def get_channel_surface(self, path):
         try:
             if self.content_type == TV_GROUP:
-                pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(path, 64, 32)
+                surface = self.get_surface_for_file(path, 64, 32)
             elif self.content_type == MOVIES_GROUP:
-                pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(path, 200, 200)
+                surface = self.get_surface_for_file(path, 200, 200)
             else:
-                pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(path, 200, 200)
+                surface = self.get_surface_for_file(path, 200, 200)
         except:
-            pixbuf = self.generic_channel_pixbuf
-        return pixbuf
+            surface = self.get_surface_for_file("/usr/share/hypnotix/generic_tv_logo.png", 22, 22)
+        return surface
 
     def on_go_back_button(self, widget):
         self.navigate_to(self.back_page)
         if self.active_channel != None:
             self.playback_bar.show()
+        if self.active_group and self.back_page == "categories_page":
+            self.init_channels_flowbox()
+
+    def on_search_button_toggled(self, widget):
+        if self.search_button.get_active():
+            self.search_bar.show()
+            self.search_bar.grab_focus()
+        else:
+            self.search_bar.hide()
+
+    def on_search_bar(self, widget):
+        self.content_type = TV_GROUP
+        search_bar_text = unidecode(self.search_bar.get_text()).lower()
+        if search_bar_text != self.latest_search_bar_text:
+            self.latest_search_bar_text = search_bar_text
+            self.search_bar.set_sensitive(False)
+            GLib.timeout_add_seconds(0.1, self.on_search)
+
+    def on_search(self):
+        def filter_func(child):
+            search_bar_text = unidecode(self.search_bar.get_text()).lower()
+            label_text = unidecode(child.get_children()[0].get_children()[0].get_children()[1].get_text()).lower()
+            if search_bar_text in label_text:
+                return True
+            else:
+                return False
+
+        self.channels_flowbox.set_filter_func(filter_func)
+        if not self.channels_flowbox.get_children():
+            self.show_channels(self.active_provider.channels)
+        self.status(None)
+        print("Filtering %d channel names containing the string '%s'..." % (len(self.channels_flowbox.get_children()), self.latest_search_bar_text))
+        self.search_bar.set_sensitive(True)
+        self.search_bar.grab_focus_without_selecting()
+        self.navigate_to("channels_page")
+
+    def init_channels_flowbox(self):
+        self.latest_search_bar_text = None
+        self.active_group = None
+        for child in self.channels_flowbox.get_children():
+            self.channels_flowbox.remove(child)
+        self.channels_flowbox.invalidate_filter()
 
     @idle_function
     def navigate_to(self, page, name=""):
         self.go_back_button.show()
+        self.search_button.show()
         self.fullscreen_button.hide()
         self.stack.set_visible_child_name(page)
         provider = self.active_provider
@@ -895,6 +972,7 @@ class MainWindow():
     def on_provider_selected(self, widget, provider):
         self.active_provider = provider
         self.settings.set_string("active-provider", provider.name)
+        self.init_channels_flowbox()
         self.navigate_to("landing_page")
 
     def on_preferences_button(self, widget):
@@ -1164,13 +1242,44 @@ class MainWindow():
         self.application.quit()
 
     def on_key_press_event(self, widget, event):
-        ctrl = (event.state & Gdk.ModifierType.CONTROL_MASK)
+        # Get any active, but not pressed modifiers, like CapsLock and NumLock
+        persistant_modifiers = Gtk.accelerator_get_default_mod_mask()
+
+        # Determine the actively pressed modifier
+        modifier = event.get_state() & persistant_modifiers
+        # Bool of Control or Shift modifier states
+        ctrl = modifier == Gdk.ModifierType.CONTROL_MASK
+        shift = modifier == Gdk.ModifierType.SHIFT_MASK
+
         if ctrl and event.keyval == Gdk.KEY_r:
             self.reload(page=None, refresh=True)
+        elif ctrl and event.keyval == Gdk.KEY_f:
+            if self.search_button.get_active():
+                self.search_button.set_active(False)
+            else:
+                self.search_button.set_active(True)
         elif event.keyval == Gdk.KEY_F11 or \
-             event.keyval == Gdk.KEY_f or \
+             (event.keyval == Gdk.KEY_f and not ctrl and type(widget.get_focus()) != gi.repository.Gtk.SearchEntry) or \
              (self.fullscreen and event.keyval == Gdk.KEY_Escape):
             self.toggle_fullscreen()
+        # elif event.keyval == Gdk.KEY_Left:
+        #     # Left of in the list
+        #     pass
+        # elif event.keyval == Gdk.KEY_Right:
+        #     # Right of in the list
+        #     pass
+        # elif event.keyval == Gdk.KEY_Up:
+        #     # Up of in the list
+        #     pass
+        # elif event.keyval == Gdk.KEY_Down:
+        #     # Down of in the list
+        #     pass
+        #elif event.keyval == Gdk.KEY_Escape:
+        #    # Go back one level
+        #    pass
+        # #elif event.keyval == Gdk.KEY_Return:
+        #     # Same as click
+        # #    pass
 
     @async_function
     def reload(self, page=None, refresh=False):
@@ -1179,23 +1288,64 @@ class MainWindow():
         for provider_info in self.settings.get_strv("providers"):
             try:
                 provider = Provider(name=None, provider_info=provider_info)
-                if refresh:
-                    self.status("Downloading playlist...", provider)
+
+                # Add provider to list. This must be done so that it shows up in the
+                # list of providers for editing.
+                self.providers.append(provider)
+
+                if provider.type_id != "xtream":
+                    # Download M3U
+                    if refresh:
+                        self.status("Downloading playlist...", provider)
+                    else:
+                        self.status("Getting playlist...", provider)
+                    ret = self.manager.get_playlist(provider, refresh=refresh)
+                    if ret:
+                        self.status("Checking playlist...", provider)
+                        if (self.manager.check_playlist(provider)):
+                            self.status("Loading channels...", provider)
+                            self.manager.load_channels(provider)
+                            if provider.name == self.settings.get_string("active-provider"):
+                                self.active_provider = provider
+                            self.status(None)
+                            print("%s: %d channels, %d groups, %d series, %d movies" % (provider.name, \
+                                len(provider.channels), len(provider.groups), len(provider.series), len(provider.movies)))
+                    else:
+                        self.status("Failed to Download playlist from {}".format(provider.name),provider)
+
                 else:
-                    self.status("Getting playlist...", provider)
-                self.manager.get_playlist(provider, refresh=refresh)
-                self.status("Checking playlist...", provider)
-                if (self.manager.check_playlist(provider)):
-                    self.status("Loading channels...", provider)
-                    self.manager.load_channels(provider)
-                    self.providers.append(provider)
-                    if provider.name == self.settings.get_string("active-provider"):
-                        self.active_provider = provider
-                    self.status(None)
+                    # Load xtream class
+                    from xtream import XTream
+                    # Download via Xtream
+                    self.x = XTream(provider.name,provider.username,provider.password,provider.url,os.path.expanduser("~/.hypnotix/providers"))
+                    if self.x.auth_data != {}:
+                        print("XTREAM `{}` Loading Channels".format(provider.name))
+                        self.x.load_iptv()
+                        # Inform Provider of data
+                        provider.channels = self.x.channels
+                        provider.movies = self.x.movies
+                        provider.series = self.x.series
+                        provider.groups = self.x.groups
+
+                        # Change redownload timeout
+                        self.reload_timeout_sec = 60*60*2 # 2 hours
+                        if self._timerid:
+                            GLib.source_remove(self._timerid)
+                        self._timerid = GLib.timeout_add_seconds(self.reload_timeout_sec, self.force_reload)
+
+                        # If no errors, approve provider
+                        if provider.name == self.settings.get_string("active-provider"):
+                            self.active_provider = provider
+                        self.status(None)
+                    else:
+                        print("XTREAM Authentication Failed")
+
             except Exception as e:
                 print(e)
                 traceback.print_exc()
                 print("Couldn't parse provider info: ", provider_info)
+
+        # If there are more than 1 providers and no Active Provider, set to the first one
         if len(self.providers) > 0 and self.active_provider == None:
             self.active_provider = self.providers[0]
 
@@ -1204,6 +1354,7 @@ class MainWindow():
         if page != None:
             self.navigate_to(page)
         self.status(None)
+        self.latest_search_bar_text = None
 
     def force_reload(self):
         self.reload(page=None, refresh=True)
@@ -1244,6 +1395,9 @@ class MainWindow():
         options["user_agent"] = self.settings.get_string("user-agent")
         options["referrer"] = self.settings.get_string("http-referer")
 
+        while not self.mpv_drawing_area.get_window() and not Gtk.events_pending():
+            time.sleep(0.1)
+
         self.mpv = mpv.MPV(**options, script_opts='osc-layout=box,osc-seekbarstyle=bar,osc-deadzonesize=0,osc-minmousemove=3', input_default_bindings=True, \
              input_vo_keyboard=True,osc=True, ytdl=True, wid=str(self.mpv_drawing_area.get_window().get_xid()))
 
